package de.ppi.psec.services;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.fest.assertions.Assertions.*;

import java.security.Permission;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import models.psec.PsecPermission;
import models.psec.PsecRole;
import models.psec.PsecUser;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import play.i18n.Lang;
import play.mvc.Http.Context;
import play.mvc.Http.Cookie;
import play.mvc.Http.Cookies;
import play.mvc.Http.Request;
import play.mvc.Http.RequestBody;
import play.mvc.Http.Response;
import play.mvc.Http.Session;
import de.ppi.psec.permissions.PsecBasicPermission;

@SuppressWarnings("boxing")
public class AuthenticationServiceTest {

    private AuthenticationService service;

    private static final String SESSION_KEY_USERID = "psecUid";
    private static final String SESSION_KEY_REMEMBERED = "psecRem";
    private static final String COOKIE_NAME = "psec_rememberMe";
    
    
    @Before
    public void setUp() {
        service = new AuthenticationServiceImpl(10, "100d", "abcdefghijklmnopqrst");
        initSessionRequestAndResponse();
    }
    
    private void initSessionRequestAndResponse() {
        final Request request = new DummyRequest();
        Context.current.set(new Context(request, new HashMap <String, String>(), 
                new HashMap <String, String>()));
        
    }

    @BeforeClass
    public static void setUpBeforeClass() throws Exception {
    }


    @Test
    public void testIsAuthenticate() {
        assertFalse(service.isAuthenticate(true));
        getSession().put(SESSION_KEY_USERID, "1");
        assertTrue(service.isAuthenticate(false));
        getSession().put(SESSION_KEY_REMEMBERED, "1");
        assertTrue(service.isAuthenticate(false));
        getSession().put(SESSION_KEY_REMEMBERED, "Y");
        assertFalse(service.isAuthenticate(false));
    }

   
    
    @Test
    public void testCanRemembered() {
        boolean[] booleans = {true, false};
        AuthenticationServiceStub serviceStub = new AuthenticationServiceStub();
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);
        serviceStub.addUser(psecUser);
        for (boolean remember : booleans) {
            for (boolean js : booleans) {
                assertTrue(serviceStub.login("hugo", "secret", remember, js));
                copyCookiesFromResponseToRequest();
                assertEquals(remember, serviceStub.canRemembered());
                initSessionRequestAndResponse();
            }
        }
        assertTrue(serviceStub.login("hugo", "secret", true, true));
        copyCookiesFromResponseToRequest();
        assertTrue(serviceStub.canRemembered());
        serviceStub.logout();
        assertNotNull(getDiscardedCookie(COOKIE_NAME));
        ((DummyRequest)getRequest()).removeCookie(COOKIE_NAME);
        assertFalse(serviceStub.canRemembered());
        
    }

    private void copyCookiesFromResponseToRequest() {
        final Iterable <Cookie> cookies = getResponse().cookies();
        DummyRequest request = (DummyRequest) getRequest();
        for (Cookie cookie : cookies) {
            request.addCookie(cookie.name(), cookie);
        }
    }
    
    @Test
    public void testLogin() {
        boolean[] booleans = {false, true};
        AuthenticationServiceStub serviceStub = new AuthenticationServiceStub();
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);
        serviceStub.addUser(psecUser);
        for (boolean remember : booleans) {
            for (boolean js : booleans) {
            	assertNull(getCookie(COOKIE_NAME));
                assertTrue(serviceStub.login("hugo", "secret", remember, js));                
                assertEquals(remember, getCookie(COOKIE_NAME) != null);
                if  (remember) {                
                    final Cookie rememberMeCookie = getCookie(COOKIE_NAME);
                    assertEquals(100*24*60*60, rememberMeCookie.maxAge());
                }
                assertEquals("1",Context.current().session().get(SESSION_KEY_USERID));
                assertFalse(Context.current().session().containsKey(SESSION_KEY_REMEMBERED));
                initSessionRequestAndResponse();
                //login failed
                //passwort wrong
                assertFalse(serviceStub.login("hugo", "secrets", remember, js));
                //loginname wrong
                assertFalse(serviceStub.login("hugos", "secret", remember, js));
                //user disabled.
                psecUser.setEnabled(false);
                assertFalse(serviceStub.login("hugos", "secret", remember, js));
                psecUser.setEnabled(true);
                
                //test failCount
                psecUser.setLoginFailedCount(10);
                assertTrue(serviceStub.login("hugo", "secret", remember, js));
                assertEquals(0, psecUser.getLoginFailedCount());
                psecUser.setLoginFailedCount(10);
                assertFalse(serviceStub.login("hugo", "secrets", remember, js));
                assertEquals(11, psecUser.getLoginFailedCount());
                assertFalse(serviceStub.login("hugo", "secret", remember, js));
                psecUser.setLoginFailedCount(0);
                initSessionRequestAndResponse();
            }
        }
    }

    private Cookie getCookie(String name) {
    	final Iterable<Cookie> cookies = getResponse().cookies();
    	for (Cookie cookie : cookies) {
			if (name.equals(cookie.name())) {
				return cookie;
			}
    	}
    	return null;
    }
    

    private String getDiscardedCookie(String name) {
        final Iterable<String> cookieNames = getResponse().discardedCookies();
        for (String cookieName: cookieNames) {
            if (name.equals(cookieName)) {
                return cookieName;
            }
        }
        return null;
    }
    
    @Test
    public void testLogout() {
        getSession().put(SESSION_KEY_USERID, "1");
        getResponse().setCookie(COOKIE_NAME, "A value");
        assertNotNull(getSession().get(SESSION_KEY_USERID));
        Cookie cookie = getCookie(COOKIE_NAME);
        assertNotNull(cookie);        
        assertThat(cookie.maxAge()).isNotEqualTo(0);
        assertEquals("A value", cookie.value());
        ((DummyRequest)getRequest()).addCookie(COOKIE_NAME, cookie);
        service.logout();
        assertNull(getSession().get(SESSION_KEY_USERID));
        assertNotNull(getDiscardedCookie(COOKIE_NAME));
    }

    
    @Test
    public void testSessionValues() {
        AuthenticationServiceStub serviceStub = new AuthenticationServiceStub();
        final PsecUser psecUser = new PsecUser();
        psecUser.setId(Long.valueOf(1));
        psecUser.setLoginname("hugo");
        psecUser.setPassword("secret");
        psecUser.setFirstname("Hugo");
        psecUser.setLastname("Boss");
        psecUser.setLoginFailedCount(0);
        psecUser.setEnabled(true);

        final List <String> roleNames = new ArrayList <String>();
        final List <String> permNames = new ArrayList <String>();
        final List <PsecPermission> permissions = new ArrayList <PsecPermission>();
        for (int i = 0; i < 10; i++) {
            final  String roleName = "role_" + i;
            final String permName = "perm_" + i;
            roleNames.add(roleName);
            permNames.add(permName);
            final PsecRole role = new PsecRole();    
            role.setName(roleName);
            final PsecPermission perm = createPermission(permName);    
            final List <PsecPermission> rolePermissions = new ArrayList <PsecPermission>();
            final PsecPermission permRole = createPermission(roleName);    
            rolePermissions.add(permRole);
            role.setPsecPermissions(rolePermissions); 
            permissions.add(perm);
            psecUser.addToRoles(role);
        }
        //TODO niels fix it psecUser.setPermissions(permissions);
        
        serviceStub.addUser(psecUser);
        assertTrue(serviceStub.login("hugo", "secret", true, true));
        
        assertEquals(psecUser.getFirstname(), serviceStub.getFirstname());
        assertEquals(psecUser.getLastname(), serviceStub.getLastname());
        assertEquals(psecUser.getLoginname(), serviceStub.getLoginname());
        assertEquals(psecUser.getId().toString(), serviceStub.getUserid());
        final List<String> serviceRoleList = serviceStub.getRoles();
        assertEquals(roleNames.size(), serviceRoleList.size());
        for (int i = 0; i < roleNames.size(); i++) {
            assertEquals(roleNames.get(i), serviceRoleList.get(i));
        }
        final List<Permission> servicePermList = serviceStub.getPermissions();
        assertEquals(permissions.size(), servicePermList.size());
        for (int i = 0; i < permissions.size(); i++) {
            assertEquals(permissions.get(i).convertToPermission(), servicePermList.get(i));
        }
    }
    
    private PsecPermission createPermission(String name) {
        PsecPermission result = new PsecPermission();
        result.setName(name);
        result.setType(PsecBasicPermission.class.getName());
        result.setActions(name);
        result.setTarget(name);
        return result;
    }
    
    private Session getSession() {
        return Context.current().session();
    }
    
    private Request getRequest() {
        return Context.current().request();
    }
    
    private Response getResponse() {
        return Context.current().response();
    }

    private final class DummyRequest extends Request {
        
        private final Map <String, String[]> headers = new HashMap <String, String[]>();
        
        private Map <String, Cookie> cookieMap = new HashMap <String, Cookie>() ; 
        
        final Cookies cookies = new Cookies() {
            
            @Override
            public Cookie get(String name) {
                return cookieMap.get(name);
            }
            
        };
        
        public void addCookie(String name, Cookie cookie) {
            cookieMap.put(name, cookie);
        }
        
        public void removeCookie(String name) {
            cookieMap.remove(name);
        }
        
        
        @Override
        public String uri() {
            return null;
        }
        
        @Override
        public Map <String, String[]> queryString() {
            return null;
        }
        
        @Override
        public String path() {
            return null;
        }
        
        @Override
        public String method() {
            return null;
        }
        
        @Override
        public String host() {
            return null;
        }
        
        @Override
        public Map <String, String[]> headers() {                
            return headers;
        }
        
        @Override
        public Cookies cookies() {
            return cookies;
        }
        
        @Override
        public RequestBody body() {
            return null;
        }
        
        @Override
        public List <Lang> acceptLanguages() {
            return null;
        }
    }
        
}
